import logging
import os
from typing import Optional

import socketio
from .http_client import HttpClient
from .socket_client import SocketClient

ACHO_TOKEN = os.environ.get("ACHO_PYTHON_SDK_TOKEN") or ""
BASE_URL = os.environ.get("ACHO_PYTHON_SDK_BASE_URL") or ""
BASE_SOCKET_NAMESPACES = ['/soc']
DEFAULT_SOCKET_NAMESPACE = '/soc'
ACHO_CLIENT_TIMEOUT = 30
APP_ENDPOINTS = 'apps'

logging.basicConfig(format='%(levelname)s: %(name)s | %(message)s', level=logging.DEBUG)

logger = logging.getLogger(__name__)

class App():
    
    # sio = socketio.AsyncClient(logger=True, engineio_logger=True)

    def __init__(self, id: str, token: Optional[str] = ACHO_TOKEN, base_url = BASE_URL, timeout = ACHO_CLIENT_TIMEOUT):
        self.http = HttpClient(token=token, base_url=base_url, timeout=timeout)
        self.app_id = id
        return

    async def versions(self):
        versions = await self.http.call_api(path=f"{APP_ENDPOINTS}/{self.app_id}/versions", http_method="GET")
        return versions
    
    async def details(self):
        details = await self.http.call_api(path=f"{APP_ENDPOINTS}/{self.app_id}", http_method="GET")
        return details
    
    def version(self, app_version_id: str):
        return AppVersion(app_id=self.app_id, app_version_id=app_version_id, token=self.http.token, base_url=self.http.base_url, timeout=self.http.timeout)
    
    async def version_editing(self):
        appDetails = await self.details()
        userDefaults = appDetails['userDefaults']
        app_version_id = userDefaults['version_id']
        return AppVersion(app_id=self.app_id, app_version_id=app_version_id, token=self.http.token, base_url=self.http.base_url, timeout=self.http.timeout)
    
    async def version_published(self):
        versions = await self.versions()
        for version in versions:
            if version['status'] in ['published']:
                return AppVersion(app_id=self.app_id, app_version_id=version['id'], token=self.http.token, base_url=self.http.base_url, timeout=self.http.timeout)
        raise('No published version found for app: ', self.app_id)
            
    def push_event(self, event: dict):
        logger.warning('Please specify version before publishing events')
        return
    
class AppVersion():

    # sio = socketio.AsyncClient(logger=True, engineio_logger=True)

    def __init__(self, app_id: str, app_version_id: str, token: Optional[str] = None, base_url = BASE_URL, socket_namespaces = BASE_SOCKET_NAMESPACES, timeout = ACHO_CLIENT_TIMEOUT):
        self.socket_url = f'{base_url}{DEFAULT_SOCKET_NAMESPACE}'
        self.socket = SocketClient(app_version_id=app_version_id, token=token, base_url=self.socket_url, socket_namespaces=socket_namespaces, timeout=timeout)
        self.http = HttpClient(token=token, base_url=base_url, timeout=timeout)
        self.app_id = app_id
        self.app_version_id = app_version_id
        self.connected = False
        self.joined = False
        return
    
    def version_default_handlers(self):
        self.socket.sio.on('connect', namespace=DEFAULT_SOCKET_NAMESPACE, handler=self.reconnect_attempt)
    
    async def connect(self, namespaces: Optional[list] = DEFAULT_SOCKET_NAMESPACE):
        try:
            await self.http.identify()
            self.socket.default_handlers()
            result = await self.socket.conn(namespaces=namespaces)
            self.connected = True
            self.version_default_handlers()
            await self.join()
            await self.nb_claim()
            return result
        except Exception as e:
            self.connected = False
            raise Exception(e)
        
    async def reconnect_attempt(self):
        logger.info('reconnecting')
        try:
            # if (self.connected):
            #     await self.connect()
            if (self.joined):
                await self.join()
                await self.nb_claim()
            return
        except Exception as e:
            raise Exception(e)

    async def join(self, namespaces: Optional[list] = DEFAULT_SOCKET_NAMESPACE):
        try:
            logger.debug({'app_version_id': self.app_version_id, 'is_editing': True})
            result = await self.socket.sio.emit('join_app_builder_room', {'app_version_id': self.app_version_id}, namespace=namespaces)
            self.joined = True
            return result
        except Exception as e:
            self.joined = False
            raise Exception(e)

    async def leave(self, namespaces: Optional[list] = DEFAULT_SOCKET_NAMESPACE):
        try:
            result = await self.socket.sio.emit('leave_app_builder_room', {'app_version_id': self.app_version_id}, namespace=namespaces)
            self.joined = False
            return result
        except Exception as e:
            self.joined = False
            raise Exception(e)
    
    async def nb_nodes(self):
        nodes = await self.http.call_api(path=f"/apps/{self.app_id}/versions/{self.app_version_id}/nb-nodes", http_method="GET")
        return nodes
    
    async def nb_claim(self):
        if (not self.connected):
            logger.warning('Please connect to socket first')
        if (not self.joined):
            logger.warning('Please join app builder room first')
        nodes = await self.nb_nodes()
        for node in nodes:
            if node['endpointUrl'] == self.socket.notebook_name:
                self.socket.node_id = node['id']
                break
        if self.socket.node_id is None:
            logger.warning('Current nodebook is not claimed by any node')
            return
        else:
            logger.info(f'Current nodebook is claimed by node: {self.socket.node_id}')
            await self.socket.notebook_detect({'app_version_id': self.app_version_id})
            return
    
    async def send_webhook(self, event: dict):
        event.update({'scope': self.app_version_id})
        event.update({'type': 'notebook_event'})
        event.update({'notebook_name': self.socket.notebook_name})
        event.update({'nodeId': self.socket.node_id})
        payload = {
            'scope': self.app_version_id,
            'event': event
        }
        logger.debug('sending webhook')
        logger.debug(payload)
        return await self.http.call_api(path="neurons/webhook", http_method="POST", json=payload)
    
    async def push_event(self, event: dict):
        event.update({'scope': self.app_version_id})
        result = await self.socket.sio.emit('push', event, namespace=DEFAULT_SOCKET_NAMESPACE)
        return result


import logging
import os
import pandas as pd
from typing import Optional

from .http_client import HttpClient

logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.DEBUG)

ACHO_TOKEN = os.environ.get("ACHO_PYTHON_SDK_TOKEN") or ""
BASE_URL = os.environ.get("ACHO_PYTHON_SDK_BASE_URL") or ""
ACHO_CLIENT_TIMEOUT = 30

class View():

    def __init__(self, http_client: HttpClient, project_id: Optional[int] = None, view_id: Optional[int] = None):
        self.http = http_client
        self.project_id = project_id
        self.view_id = view_id
        return
    
    def __str__(self) -> str:
        return f"View {self.view_id} of Project {self.project_id}"
    
    async def download(self, options = {}, destination = os.getcwd()):
        payload = {
            'proj_id': self.project_id,
            'view_id': self.view_id,
            'options': options,
        }
        resp = await self.http.call_api(path=f"/project/get-csv-links", http_method="POST", json=payload)
        if not os.path.exists(destination):
            os.makedirs(destination)
        # destination_path = os.path.join(destination, filename)
        result = resp.get('signedUrls', [])
        return result
    
    async def export_to_df(self, options = {}):
        urls = await self.download(options)
        aggregated_results = pd.DataFrame()

        for url in urls:
            # Adjust the chunksize parameter as needed to control memory usage
            for chunk in pd.read_json(url, lines=True, compression='gzip', chunksize=10000):
                # Process the chunk: perform operations like filtering, aggregations, etc.
                # This example assumes a simple operation, but you would replace this with whatever operation you need
                processed_chunk = chunk # This is where you'd apply operations

                # Update aggregated results with processed chunk
                # This might involve concatenation, summation, or any other form of aggregation depending on your need
                aggregated_results = pd.concat([aggregated_results, processed_chunk])

        return aggregated_results



import platform
import sys
from typing import Any, Optional, Union, Dict
from urllib.parse import urljoin
from acho import version

def _get_url(base_url: str, path: str) -> str:
    """Join the base_url with path
    """
    return urljoin(base_url, path)

def _build_req_args(
    *,
    token: Optional[str],
    http_method: str,
    default_params: dict,
    params: dict,
    json: dict,  # skipcq: PYL-W0621
    headers: dict,
    auth: dict,
) -> dict:
    has_json = json is not None
    if has_json and http_method != "POST":
        msg = "Json data can only be submitted as POST requests. GET requests should use the 'params' argument."
        raise TypeError(msg)

    if params is not None and isinstance(params, dict):
        params = {k: v for k, v in params.items() if v is not None}
        _set_default_params(params, default_params)
    if json is not None and isinstance(json, dict):
        _set_default_params(json, default_params)

    token: Optional[str] = token
    if params is not None and "token" in params:
        token = params.pop("token")
    if json is not None and "token" in json:
        token = json.pop("token")
    req_args = {
        "headers": _get_headers(
            headers=headers,
            token=token,
            has_json=has_json,
            request_specific_headers=headers,
        ),
        "params": params,
        "json": json,
        "auth": auth,
    }
    return req_args


def _set_default_params(target: dict, default_params: dict) -> None:
    for name, value in default_params.items():
        if name not in target:
            target[name] = value

def _get_headers(
    *,
    headers: dict,
    token: Optional[str],
    has_json: bool,
    request_specific_headers: Optional[dict],
) -> Dict[str, str]:
    """Constructs the headers need for a request.
    Args:
        has_json (bool): Whether or not the request has json.
        has_files (bool): Whether or not the request has files.
        request_specific_headers (dict): Additional headers specified by the user for a specific request.
    Returns:
        The headers dictionary.
            e.g. {
                'Content-Type': 'application/json;charset=utf-8',
                'Authorization': 'jwt ey.....',
                'User-Agent': 'Python/3.10.6 Acho/0.0.1 Darwin/17.7.0'
            }
    """
    final_headers = {
        "Content-Type": "application/x-www-form-urlencoded",
    }
    if headers is None or "User-Agent" not in headers:
        final_headers["User-Agent"] = get_user_agent()

    if token:
        final_headers.update({"Authorization": "jwt {}".format(token)})
    if headers is None:
        headers = {}

    # Merge headers specified at client initialization.
    final_headers.update(headers)

    # Merge headers specified for a specific request. e.g. oauth.access
    if request_specific_headers:
        final_headers.update(request_specific_headers)

    if has_json:
        final_headers.update({"Content-Type": "application/json;charset=utf-8"})

    # if has_files:
    #     # These are set automatically by the aiohttp library.
    #     final_headers.pop("Content-Type", None)

    return final_headers

def get_user_agent(prefix: Optional[str] = None, suffix: Optional[str] = None):
    """Construct the user-agent header with the package info,
    Python version and OS version.
    Returns:
        The user agent string.
        e.g. 'Python/3.6.7 acho_py/0.0.1 Darwin/17.7.0'
    """
    # __name__ returns all classes, we only want the client
    client = "{0}/{1}".format("acho_py", version.__version__)
    python_version = "Python/{v.major}.{v.minor}.{v.micro}".format(v=sys.version_info)
    system_info = "{0}/{1}".format(platform.system(), platform.release())
    user_agent_string = " ".join([python_version, client, system_info])
    prefix = f"{prefix} " if prefix else ""
    suffix = f" {suffix}" if suffix else ""
    return prefix + user_agent_string + suffix
    


import asyncio
import logging
import os
from typing import Optional

from .http_client import HttpClient

logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.DEBUG)

ACHO_TOKEN = os.environ.get("ACHO_PYTHON_SDK_TOKEN") or ""
BASE_URL = os.environ.get("ACHO_PYTHON_SDK_BASE_URL") or ""
ACHO_CLIENT_TIMEOUT = 30

class Acho():

    def __init__(self, token: Optional[str] = ACHO_TOKEN, base_url = BASE_URL, timeout = ACHO_CLIENT_TIMEOUT):
        self.http = HttpClient(token=token, base_url=base_url, timeout=timeout)
        return
    
    def __str__(self):
        return "Acho Python SDK"
    

class AssetManager():

    def __init__(self, path = '', token: Optional[str] = ACHO_TOKEN, base_url = BASE_URL, timeout = ACHO_CLIENT_TIMEOUT):
        self.http = HttpClient(token=token, base_url=base_url, timeout=timeout)
        self.path = path
        return

    def redirect(self, path):
        self.path = path

    async def list(self, options = {}):
        payload = {
            'path': self.path,
            'depth': options.get('depth', 1),
            'options': options,
        }
        result = await self.http.call_api(path=f"/uploaded/search", http_method="POST", json=payload)
        return result

    async def upload(self, options = {}):
        params = {
            'destination': self.path,
            'public': options.get('public', 'true'),
        }
        # TODO: Add files to multipart/form-data
        result = await self.http.call_api(path=f"/uploaded/file/upload", http_method="POST", params=params)
        return result
    
    async def download(self, filename, options = {}, destination = os.getcwd()):
        payload = {
            'file': { 'path': os.path.join(self.path, filename) },
            'options': options,
        }
        # Save file stream to storage / memory
        http_stream = await self.http.stream_api(path=f"/uploaded/file/download", http_method="POST", json=payload)
        if not os.path.exists(destination):
            os.makedirs(destination)
        destination_path = os.path.join(destination, filename)
        with open(destination_path, 'wb') as f:
            for chunk in http_stream.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
        return destination_path

    async def get_url(self, filename, options = {}):
        payload = {
            'file': { 'path': self.path + filename },
            'options': options,
        }
        result = await self.http.call_api(path=f"/uploaded/file/get-link", http_method="POST", json=payload)
        return result




import logging
import os
from typing import Optional
import socketio
import ipynbname

logger = logging.getLogger(__name__)

class SocketClient:
    BASE_URL = os.environ.get("ACHO_PYTHON_SDK_BASE_URL") or ""
    BASE_SOCKET_NAMESPACES = ['/soc']
    sio = socketio.AsyncClient(logger=False, engineio_logger=False)

    def __init__(self, app_version_id: str, token: Optional[str] = None, base_url = BASE_URL, socket_namespaces = BASE_SOCKET_NAMESPACES, sio = sio, timeout = 30, notebook_name = 'unknown_notebook'):
        self.token = None if token is None else token.strip()
        """A JWT Token"""
        self.base_url = base_url
        self.socket_namespaces = socket_namespaces
        self.default_namespace = '/soc'
        """A string representing the Acho API base URL.
        Default is `'https://kube.acho.io'`."""
        self.timeout = timeout
        """The maximum number of seconds client staying alive"""
        self.default_params = {}
        self.sio = sio
        self.notebook_name = notebook_name
        self.node_id = None
        self.app_version_id = app_version_id
        self.get_notebook_attr()

    async def conn(self, namespaces: Optional[list] = None):
        logger.debug(f"Connecting to namespace {namespaces or self.socket_namespaces}")
        try:
            authenticated_url = f'{self.base_url}?token=jwt {self.token}'
            logger.debug(f'authenticated_url: {authenticated_url}')
            logger.debug(f'status: {self.sio.connected}')
            if (self.sio.connected):
                logger.debug('already connected')
                return
            else:
                result = await self.sio.connect(url=authenticated_url, namespaces=namespaces or self.socket_namespaces)
                return result
        except Exception as e:
            logger.error(f"Socket connection failed with error {e}")
            raise Exception(f"Socket connection failed with error {e}")

    def get_notebook_attr(self):
        try:
            ip = get_ipython()
            hm = ip.history_manager
            ipsession = hm.get_session_info()
            dt = ipsession[1]
            dt_string = dt.strftime("%s")
            ipname = ipynbname.name()
            ipsession = hm.get_session_info()
            self.notebook_name = ipname
            self.started_at = dt_string
            return
        except Exception as e:
            logger.error('Please run this command in a notebook')

    def default_handlers(self):
        self.sio.on('notebook_detect', namespace=self.default_namespace, handler=self.notebook_detect)
        self.sio.on('notebook_claim', namespace=self.default_namespace, handler=self.notebook_claim)

    def hook(self, event: str, callback):
        self.sio.on(event, namespace=self.default_namespace, handler=callback)

    # async def event_parser(self, callback):
    #     return lambda data: (await callback(data) for _ in '_').__anext__()
        
    async def notebook_detect(self, data):
        logger.debug('notebook detection request')
        print('notebook detection request')
        app_version_id = data['app_version_id']
        detect_result = await self.sio.emit('notebook_ready', data={'app_version_id': app_version_id, 'nodeId': self.node_id, 'notebook_name': self.notebook_name}, namespace=self.default_namespace)

    async def notebook_claim(self, data):
        logger.debug('notebook claim request')
        print('notebook claim request')
        app_version_id = data['app_version_id']
        notebook_name = data['notebook_name']
        node_id = data['nodeId']
        if (notebook_name == self.notebook_name):
            self.node_id = node_id
            if (node_id):
                detect_result = await self.sio.emit('notebook_claimed', data={'app_version_id': app_version_id, 'nodeId': node_id, 'notebook_name': self.notebook_name}, namespace=self.default_namespace)
            else:
                await self.sio.emit('notebook_claim_failed', data={'app_version_id': app_version_id, 'nodeId': node_id, 'notebook_name': self.notebook_name}, namespace=self.default_namespace)
                logger.warn('node_id is missing')

    def tag(self):
        return f'[{self.app_version_id[-4:]}][{self.node_id[-4:]}]'

    @sio.on('connect', namespace='/soc')
    def on_connect():
        # logger.debug("I'm connected to the /soc namespace!")
        print("I'm connected to the /soc namespace!")
        return
    
    @sio.event
    async def connect():
        logger.info('connected to server')
        return
    
    @sio.event
    async def connect_error(data):
        print("The connection failed!")

    @sio.event
    async def disconnect():
        logger.info('disconnected from server')
        return

    @sio.on('*', namespace='/soc')
    async def catch_all(event, data):
        logger.info(f'incoming event: {event}')
        return

    


import asyncio
import logging
import os
import pandas as pd
from typing import Optional

from .http_client import HttpClient
from .view_client import View

logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.DEBUG)

ACHO_TOKEN = os.environ.get("ACHO_PYTHON_SDK_TOKEN") or ""
BASE_URL = os.environ.get("ACHO_PYTHON_SDK_BASE_URL") or ""
ACHO_CLIENT_TIMEOUT = 30

class Project():

    def __init__(self, token: Optional[str] = ACHO_TOKEN, base_url = BASE_URL, timeout = ACHO_CLIENT_TIMEOUT):
        self.http = HttpClient(token=token, base_url=base_url, timeout=timeout)
        self.project_id = None
        return
    
    def __str__(self) -> str:
        return f"Acho Python SDK Project Client"
    
    def set_current(self, project_id):
        self.project_id = project_id
        return
    
    async def list(self, options = {}):
        data = await self.http.call_api(path=f"/project/list", http_method="POST", json={})
        result = data.get('result', [])
        df_projects = pd.json_normalize(result)
        return df_projects
    
    async def list_views(self, options = {}):
        if self.project_id is None or not isinstance(self.project_id, int):
            try:
                self.project_id = int(self.project_id)
            except (ValueError, TypeError):
                logging.warning(f"Project ID is None or not resolvable to an integer: {self.project_id}, use set_current() to set a project ID")
            return
        data = await self.http.call_api(path=f"/project/{self.project_id}/views", http_method="GET")
        result = data.get('projectViewList', [])
        df_views = pd.json_normalize(result)
        return df_views
    
    async def view(self, view_id, options = {}):
        if self.project_id is None or not isinstance(self.project_id, int):
            try:
                self.project_id = int(self.project_id)
            except (ValueError, TypeError):
                logging.warning(f"Project ID is None or not resolvable to an integer: {self.project_id}, use set_current() to set a project ID")
            return
        
        if view_id is None or not isinstance(view_id, int):
            try:
                view_id = int(self.project_id)
            except (ValueError, TypeError):
                logging.warning(f"View ID is None or not resolvable to an integer: {view_id}")
            return
        
        view = View(http_client=self.http, project_id=self.project_id, view_id=view_id)
        return view



from .http_client import HttpClient
from .app_client import App
from .socket_client import SocketClient
from .studio_client import AssetManager
from .project_client import Project

__all__ = [
    "HttpClient",
    "SocketClient",
    "App",
    "AssetManager",
    "Project",
]


import asyncio
from .socket_client import SocketClient

class AppSocket(SocketClient):

    def __init__(self, token: str, app_version_id: str, base_url = SocketClient.BASE_URL, socket_namespaces = SocketClient.BASE_SOCKET_NAMESPACES, sio = SocketClient.sio, timeout = SocketClient.timeout):
        super().__init__(token=token, base_url=base_url, socket_namespaces=socket_namespaces, sio=sio, timeout=timeout)
        self.app_version_id = app_version_id
        return

    def init(self, app_version_id: str):
        self.app_version_id = app_version_id
        return


import os
import asyncio
import requests
import socketio
import urllib

from typing import Optional, Union, Dict, Any, List

from .client_utils import _build_req_args, _get_url

# TODO: Add support for multipart/form-data

class HttpClient:
    BASE_URL = os.environ.get("ACHO_PYTHON_SDK_BASE_URL") or ""

    def __init__(self, token: Optional[str] = None, base_url = BASE_URL, timeout = 30):
        self.token = None if token is None else token.strip()
        """A JWT Token"""
        self.base_url = base_url
        """A string representing the Acho API base URL.
        Default is `'https://kube.acho.io'`."""
        self.timeout = timeout
        """The maximum number of seconds client staying alive"""
        self.default_params = {}

    async def identify(self):
        return await self.call_api(path="/auth/identify", http_method="GET")

    async def call_api(self, path, http_method: str = "POST", params: Optional[dict] = None,
        json: Optional[dict] = None,
        headers: Optional[dict] = None,
        auth: Optional[dict] = None) -> Any:

        api_url = _get_url(self.base_url, path)

        if auth is not None:
            if isinstance(auth, dict):
                if headers is None:
                    headers = {}
                headers["Authorization"] = '{} {}'.format(auth.token_type, auth.token)
        else:
            headers = {}
            headers["Content-Type"] = 'application/json;charset=utf-8'
            headers["Authorization"] = 'jwt {}'.format(self.token)

        headers = headers or {}
        
        req_args = _build_req_args(
            token=self.token,
            http_method=http_method,
            default_params=self.default_params,
            params=params,
            json=json,
            headers=headers,
            auth=auth,
        )

        req_result = await self._send(
            http_method=http_method,
            api_url=api_url,
            req_args=req_args,
        )
        response = req_result[0]
        metadata = req_result[1]
        if response.status_code >= 400:
            raise Exception(f"Request failed with status code {response.status_code} and message {response.text}")
        else:
            try:
                return response.json()
            except Exception as e:
                return response.text
            
    async def stream_api(self, path, http_method: str = "POST", params: Optional[dict] = None,
        json: Optional[dict] = None,
        headers: Optional[dict] = None,
        auth: Optional[dict] = None) -> Any:

        api_url = _get_url(self.base_url, path)

        if auth is not None:
            if isinstance(auth, dict):
                if headers is None:
                    headers = {}
                headers["Authorization"] = '{} {}'.format(auth.token_type, auth.token)
        else:
            headers = {}
            headers["Content-Type"] = 'application/json;charset=utf-8'
            headers["Authorization"] = 'jwt {}'.format(self.token)

        headers = headers or {}
        
        req_args = _build_req_args(
            token=self.token,
            http_method=http_method,
            default_params=self.default_params,
            params=params,
            json=json,
            headers=headers,
            auth=auth,
        )
        if (http_method == "GET"):
            stream = requests.get(url=api_url, headers=req_args.get('headers'), params=req_args.get('params'), stream=True)
            return stream
        elif (http_method == "POST"):
            stream = requests.post(url=api_url, headers=req_args.get('headers'), params=req_args.get('params'), json=req_args.get('json'), stream=True)
            return stream
        else:
            raise Exception("Unsupported HTTP method")
        

    async def _send(self, http_method: str, api_url: str, req_args: dict) -> Any:
        """Sends the request out for transmission.
        Args:
            http_verb (str): The HTTP verb. e.g. 'GET' or 'POST'.
            api_url (str): The Acho API url'
            req_args (dict): The request arguments to be attached to the request.
            e.g.
            {
                json: {
                    'attachments': [{"pretext": "pre-hello", "text": "text-world"}],
                    'channel': '#random'
                }
            }
        """

        res = {}
        
        # print(req_args)
        # print(http_method)

        if (http_method == "GET"):
            res = requests.get(url=api_url, headers=req_args.get('headers'), params=req_args.get('params'))
        elif (http_method == "POST"):
            res = requests.post(url=api_url, headers=req_args.get('headers'), params=req_args.get('params'), json=req_args.get('json'))
        data = {
            "client": self,
            "http_method": http_method,
            "api_url": api_url,
            "req_args": req_args,
        }
        return (res, data)


from .client import HttpClient
from .client import SocketClient
from .client import App
from .client import AssetManager
from .client import Project

__all__ = ["HttpClient", "SocketClient", "App", "AssetManager", "Project"]